int test_a(void);
int test_b(void);
int test_c(void);
int test_d(void);
int test_e(void);
int test_f(void);

int main(int argc, char* argv[])
{
  test_a();
  test_b();
  test_c();
  test_d();
  test_e();
  test_f();
  return 0;
}
