
#ifndef TEST_DEFINITION
#  error Expected TEST_DEFINITION
#endif

#ifndef LANG_CXX
#  error Expected LANG_CXX
#endif

#ifdef UNPEXTED_DEFINITION
#  error Unexpected UNPEXTED_DEFINITION
#endif

int main(void)
{
  return 0;
}
