class HelloWorld3
{
    public static void main(String args[])
    {
        D d;
        d = new D();
        d.printName();

        E e;
        e = new E();
        e.printName();

        System.out.println("Hello World!");
    }
}
