#ifndef InIncludes_SubOwnDot_HPP
#define InIncludes_SubOwnDot_HPP

#include <QObject>

namespace InIncludes {

class SubOwnDot : public QObject
{
  Q_OBJECT
public:
  SubOwnDot();
  ~SubOwnDot();
};
}

#endif
