#ifndef OtherUnderscoreSub_P_HPP
#define OtherUnderscoreSub_P_HPP

#include <QObject>

class OtherUnderscoreSubPrivate : public QObject
{
  Q_OBJECT
public:
  OtherUnderscoreSubPrivate();
  ~OtherUnderscoreSubPrivate();
};

#endif
