#include "Gadget.hpp"
#include "Object.hpp"
#include "Object1Aliased.hpp"
#include "Object2Aliased.hpp"

int main(int argv, char** args)
{
  Gadget gadget;
  Object object;
  Object1Aliased object1Aliased;
  Object2Aliased object2Aliased;
  return 0;
}
