#include "item.hpp"
#include "sda/sda.hpp"
#include "sdb/sdb.hpp"
#include "sdc/sdc.hpp"

int main(int argv, char** args)
{
  // Object instances
  Item item;
  // Library calls
  sda();
  sdb();
  sdc();
  return 0;
}
