import "oaidl.idl";
import "ocidl.idl";

[
  object,
  uuid(258CCEBE-8EE4-4A48-B78C-AC53BCD59E28),
  dual,
  nonextensible,
  helpstring("ICMakeTest Interface"),
  pointer_default(unique)
]
interface ICMakeMidlTest : IUnknown
{
  [id(1), helpstring("method Method")] HRESULT Method();
}

[
  uuid(0537BA59-7EEC-48F8-BD4B-369BC7D9807E),
]
library CMakeMidlTestLib
{
  [
    uuid(D2A90807-019A-46E5-BF47-FF4FA4352D2A),
    helpstring("CMakeMidlTest Class")
  ]
  coclass CMakeMidlTest
  {
    [default] interface ICMakeMidlTest;
  };
}
