
#ifndef TESTSHAREDLIBDEPENDS_H
#define TESTSHAREDLIBDEPENDS_H

#include "renamed.h"
#include "testSharedLibRequired.h"
#include "testsharedlibdepends_export.h"

struct TESTSHAREDLIBDEPENDS_EXPORT TestSharedLibDepends
{
  int foo();
};

#endif
