
VERSION		1.23

HEAPSIZE	1048576,4096

EXPORTS
    cm_zlib_adler32                        @1
    cm_zlib_compress                       @2
    cm_zlib_crc32                          @3
    cm_zlib_deflate                        @4
    cm_zlib_deflateCopy                    @5
    cm_zlib_deflateEnd                     @6
    cm_zlib_deflateInit2_                  @7
    cm_zlib_deflateInit_                   @8
    cm_zlib_deflateParams                  @9
    cm_zlib_deflateReset                   @10
    cm_zlib_deflateSetDictionary           @11
    cm_zlib_gzclose                        @12
    cm_zlib_gzdopen                        @13
    cm_zlib_gzerror                        @14
    cm_zlib_gzflush                        @15
    cm_zlib_gzopen                         @16
    cm_zlib_gzread                         @17
    cm_zlib_gzwrite                        @18
    cm_zlib_inflate                        @19
    cm_zlib_inflateEnd                     @20
    cm_zlib_inflateInit2_                  @21
    cm_zlib_inflateInit_                   @22
    cm_zlib_inflateReset                   @23
    cm_zlib_inflateSetDictionary           @24
    cm_zlib_inflateSync                    @25
    cm_zlib_uncompress                     @26
    cm_zlib_zlibVersion                    @27
    cm_zlib_gzprintf                       @28
    cm_zlib_gzputc                         @29
    cm_zlib_gzgetc                         @30
    cm_zlib_gzseek                         @31
    cm_zlib_gzrewind                       @32
    cm_zlib_gztell                         @33
    cm_zlib_gzeof                          @34
    cm_zlib_gzsetparams                    @35
    cm_zlib_zError                         @36
    cm_zlib_inflateSyncPoint               @37
    cm_zlib_get_crc_table                  @38
    cm_zlib_compress2                      @39
    cm_zlib_gzputs                         @40
    cm_zlib_gzgets                         @41
