#ifndef OtherUnderscoreEXTRA_P_HPP
#define OtherUnderscoreEXTRA_P_HPP

#include <QObject>

class OtherUnderscoreExtraPrivate : public QObject
{
  Q_OBJECT
public:
  OtherUnderscoreExtraPrivate();
  ~OtherUnderscoreExtraPrivate();
};

#endif
