
#ifndef TESTSHAREDLIBREQUIREDUSER2_H
#define TESTSHAREDLIBREQUIREDUSER2_H

#include "testSharedLibRequired.h"
#include "testsharedlibrequireduser2_export.h"

struct TESTSHAREDLIBREQUIREDUSER2_EXPORT TestSharedLibRequiredUser2
{
  TestSharedLibRequired foo();
};

#endif
