
#include "arguments.h"
#include "bang.h"
#include "bar.h"
#include "bat.h"
#include "baz.h"
#include "bing.h"
#include "bung.h"
#include "common.h"
#include "foo.h"
#include "list.h"
#include "prefix_foo_bar_bat.h"
#include "target1.h"
#include "target2.h"
#include "ting.h"

int main(int, char**)
{
  return 0;
}
