/* Distributed under the OSI-approved BSD 3-Clause License.  See accompanying
   file Copyright.txt or https://cmake.org/licensing for details.  */
#ifndef YAF_H
#define YAF_H

class Yaf
{
public:
  Yaf();

public:
  void doYaf();
};

#endif
